
module decrypter_pipe(data,clock,reset,expanded_key,round_num,out_data,round_over);
input reg [7:0] data [3:0][3:0];  //INPUT DATA
input clock;
input reg reset;

input reg [7:0] expanded_key [3:0][3:0]; //KEY
input reg [3:0] round_num; //ROUND NUMBER
output reg [7:0] out_data [3:0][3:0];  //OUTPUT
output reg round_over;

parameter init1 = 4'b0000, init2 = 4'b0001, inv_shift_rows = 4'b0010, inv_sub_bytes = 4'b0011, wait_state = 4'b0100, latch_lut_data = 4'b0101, add_round_key = 4'b0110, round4a = 4'b0111, round4b = 4'd8, round4c = 4'd9, round4d = 4'd10,round4e = 4'd11,done1 = 4'd12, init3 = 4'd13, init4 = 4'd14, init5 = 4'd15, done2 = 5'd16, dummy1 = 5'd17, dummy2 = 5'd18, dummy3 = 5'd19, dummy4 = 5'd20, dummy5 = 5'd21;
reg [7:0] addr [3:0];
reg [7:0] data_lookup [3:0];
reg enable;
reg [5:0] state;
reg [7:0] data_shift_rows [3:0][3:0];
reg [7:0] inv_sub_data [3:0][3:0];
reg [7:0] round_key_out [3:0][3:0];
reg [3:0] col1_count, col2_count, col3_count, col4_count;
reg [3:0] i,j,k,l;
reg r_done;

reg [7:0] data_latch    [3:0][3:0];

reg [7:0] mix_matrix    [3:0][3:0];
reg [7:0] state_matrix4 [3:0][3:0];
reg [7:0] state_matrix4a [3:0][3:0][3:0];
reg [7:0] state_matrix4b [3:0][3:0][3:0];
reg [7:0] state_matrix4c [3:0][3:0][3:0];
reg [7:0] state_matrix4d [3:0][3:0][3:0];
reg [7:0] state_matrix4e [3:0][3:0][3:0];


//LOOK UP TABLE FOR SUBSTITUIN USING THE INVERSE S BOX
look_up_decrypt L0(
			.clk(clock),
			.enable(enable),
			.in(addr[0]),
			.out(data_lookup[0])
			 );

look_up_decrypt L1(
			.clk(clock),
			.enable(enable),
			.in(addr[1]),
			.out(data_lookup[1])
			 );

look_up_decrypt L2(
			.clk(clock),
			.enable(enable),
			.in(addr[2]),
			.out(data_lookup[2])
			 );
			 
look_up_decrypt L3(
			.clk(clock),
			.enable(enable),
			.in(addr[3]),
			.out(data_lookup[3])
			 );
			 


always@(posedge clock) begin
	if(reset) begin
		enable<=0;
		col1_count <= 0;
		col2_count <= 0;
		col3_count <= 0;
		col4_count <= 0;
		r_done<=0;
		state <= init1;
		for(i=0;i<4;i++) begin
		 for(j=0;j<4;j++) begin
		  data_shift_rows[i][j] <= 0;
		  state_matrix4[i][j] <= 0;
		  inv_sub_data[i][j] <= 0;
		  round_key_out[i][j] <= 0;
		  data_latch[i][j] <= data[i][j];
			for(k=0;k<4;k=k+1) begin
				state_matrix4a [i][j][k] <= 0;
				state_matrix4b [i][j][k] <= 0;
				state_matrix4c [i][j][k] <= 0;
				state_matrix4d [i][j][k] <= 0;
				state_matrix4e [i][j][k] <= 0;
			end
		 end
		end
		
//MIX MATRIX	    
		mix_matrix[0][0] <= 8'd14;
		mix_matrix[0][1] <= 8'd11;
		mix_matrix[0][2] <= 8'd13;
		mix_matrix[0][3] <= 8'd9;
		
		mix_matrix[1][0] <= 8'd9;
		mix_matrix[1][1] <= 8'd14;
		mix_matrix[1][2] <= 8'd11;
		mix_matrix[1][3] <= 8'd13;
		
		mix_matrix[2][0] <= 8'd13;
		mix_matrix[2][1] <= 8'd9;
		mix_matrix[2][2] <= 8'd14;
		mix_matrix[2][3] <= 8'd11;
		
		mix_matrix[3][0] <= 8'd11;
		mix_matrix[3][1] <= 8'd13;
		mix_matrix[3][2] <= 8'd9;
		mix_matrix[3][3] <= 8'd14;
	end
	else begin
	enable <= 1;
		case(state)
//EXTRA BUFFERING TO AVOID TIMING INCONSISTENCIES
		    init1:
			state <= init2;
			
		   init2:
			state <= init3;
			
		   init3:
			state <= init4;
			
			init4:
			state <= init5;
			
			init5:
			state <= inv_shift_rows;

//STEP 1 AND 2 COMBINED
			inv_shift_rows: begin
			                     for(j=0;j<4;j++) begin
	                               data_shift_rows[0][j] <= data_latch[0][j]; //first row;
							      end
							  
								data_shift_rows[1][0] <= data_latch[1][3];//second row;
								data_shift_rows[1][1] <= data_latch[1][0];
								data_shift_rows[1][2] <= data_latch[1][1];
								data_shift_rows[1][3] <= data_latch[1][2];
								
								data_shift_rows[2][0] <= data_latch[2][2];//third row
								data_shift_rows[2][1] <= data_latch[2][3];
								data_shift_rows[2][2] <= data_latch[2][0];
								data_shift_rows[2][3] <= data_latch[2][1];
								
								data_shift_rows[3][0] <= data_latch[3][1];//fourth row
								data_shift_rows[3][1] <= data_latch[3][2];
								data_shift_rows[3][2] <= data_latch[3][3];
								data_shift_rows[3][3] <= data_latch[3][0];
								
								state <= inv_sub_bytes;
							end
			
			inv_sub_bytes: begin
			                
							 if(col1_count < 4) begin
							   addr[0] <= data_shift_rows[col1_count][0];
							   col1_count <= col1_count + 1;
							   state <= wait_state;
							  end
							  else begin
							   state <= add_round_key;
							   col1_count <= 0;
							  end
							  
							 if(col2_count < 4) begin
							   addr[1] <= data_shift_rows[col2_count][1];
							   col2_count <= col2_count + 1;
							   state <= wait_state;
							  end
							  else begin
							   state <= add_round_key;
							   col2_count <= 0;
							  end
							  
							 if(col3_count < 4) begin
							   addr[2] <= data_shift_rows[col3_count][2];
							   col3_count <= col3_count + 1;
							   state <= wait_state;
							  end
							  else begin
							   state <= add_round_key;
							   col3_count <= 0;
							  end
							  
							 if(col4_count < 4) begin
							   addr[3] <= data_shift_rows[col4_count][3];
							   col4_count <= col4_count + 1;
							   state <= wait_state;
							  end
							  else begin
							   state <= add_round_key;
							   col4_count <= 0;
							  end
							 
			               end
						   
			wait_state: begin
			               state <= latch_lut_data;
			            end
						
			latch_lut_data:begin
							 inv_sub_data[col1_count-1][0] <= data_lookup[0];
							 inv_sub_data[col2_count-1][1] <= data_lookup[1];
							 inv_sub_data[col3_count-1][2] <= data_lookup[2];
							 inv_sub_data[col4_count-1][3] <= data_lookup[3];
							 state <= inv_sub_bytes;
			             end

//STEP 3: XOR WITH KEY					 
		    add_round_key: begin

							
							for(i=0;i<4;i=i+1) begin
								for(j=0;j<4;j=j+1) begin
									round_key_out[i][j] <= inv_sub_data[i][j]^expanded_key[i][j];
								end
							end
							if(round_num == 4'd10) begin
								state <= dummy1;
							end
							else begin
								state <=round4a;
							end
			          end
//IF ROUND 10, GO THROUGH THESE DUMMY STAGES TO ENSURE ALL ROUNDS HAVE SAME EXECUTION TIME			
			dummy1 : begin
				state <= dummy2;
			end
			
			dummy2 : begin
				state <= dummy3;
			end
			
			dummy3 : begin
				state <= dummy4;
			end
			
			dummy4 : begin
				state <= dummy5;
			end
			
			dummy5 : begin
				for(i=0;i<4;i=i+1) begin
					for(j=0;j<4;j=j+1) begin
						state_matrix4[i][j] <= round_key_out[i][j];
					end
				end
				state <= done1;
			end
//ROUND 4 : MULTIPLYING WITH MIX MATRIX						   
			round4a: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					for(k=0;k<4;k=k+1) begin
						if(mix_matrix[i][k]== 4'hd || mix_matrix[i][k] == 4'h9) begin
							state_matrix4a[i][j][k] <= round_key_out[k][j];
						end
						else if(mix_matrix[i][k] == 4'he) begin
							if(round_key_out[k][j] < 8'd128) begin
								state_matrix4a[i][j][k] <= round_key_out[k][j]<<1;
							end
							else begin
								state_matrix4a[i][j][k] <= ((round_key_out[k][j])<<1)^(8'd27);
							end
						end
						else if(mix_matrix[i][k] == 4'hb) begin
							if(round_key_out[k][j] <8'd128) begin
								state_matrix4a[i][j][k] <= ((round_key_out[k][j])<<1)^(round_key_out[k][j]);
							end
							else begin
								state_matrix4a[i][j][k] <= ((round_key_out[k][j])<<1)^(8'd27)^(round_key_out[k][j]);
							end							
						end
						

						if(round_key_out[k][j] < 8'd128) begin
							state_matrix4b[i][j][k] <= round_key_out[k][j]<<1;
						end
						else begin
							state_matrix4b[i][j][k] <= ((round_key_out[k][j])<<1)^(8'd27);
						end

							
					end
				end
			end
			
			
			state <= round4b;
		end
		
		round4b: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					for(k=0;k<4;k=k+1) begin
						if(state_matrix4b[i][j][k] < 8'd128) begin
							state_matrix4c[i][j][k] <= state_matrix4b[i][j][k]<<1;
						end
						else begin
							state_matrix4c[i][j][k] <= ((state_matrix4b[i][j][k])<<1)^(8'd27);
						end
					end
				end
			end
			state <= round4c;
		end
			
		round4c: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					for(k=0;k<4;k=k+1) begin
						if(state_matrix4c[i][j][k] < 8'd128) begin
							state_matrix4d[i][j][k] <= state_matrix4c[i][j][k]<<1;
						end
						else begin
							state_matrix4d[i][j][k] <= ((state_matrix4c[i][j][k])<<1)^(8'd27);
						end
					end
				end
			end
			state <= round4d;
		end
		
		round4d: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					for(k=0;k<4;k=k+1) begin
						if(mix_matrix[i][k] == 4'he || mix_matrix[i][k] == 4'hd) begin
							state_matrix4e[i][j][k] = state_matrix4a[i][j][k]^state_matrix4c[i][j][k]^state_matrix4d[i][j][k];
						end
						else if(mix_matrix[i][k] == 4'hb || mix_matrix[i][k] == 4'h9)begin
							state_matrix4e[i][j][k] = state_matrix4a[i][j][k]^state_matrix4d[i][j][k];
						end
					end
				end
			end
			state <= round4e;
		end
							

		
		
		round4e: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix4[i][j] <= state_matrix4e[i][j][0]^state_matrix4e[i][j][1]^state_matrix4e[i][j][2]^state_matrix4e[i][j][3];
				end
			end
			state <= done1;
		end			   
		    
			
			
			done1: begin
				r_done <= 1;
				state <= done2;
			end
			
			done2: begin
				r_done <= 0;
			end
		endcase
	end
end

assign out_data = state_matrix4;
assign round_over = r_done;

endmodule